@@ -16,6 +16,7 @@ UPB_PACKET = 2
 PIM_TO_BE_READY = 3
 
 
+# pylint: disable=too-few-public-methods
 class _Packet:
     """Details about a packet being sent"""
 
@@ -33,9 +34,10 @@ class _Packet:
 class Connection(asyncio.Protocol):
     """asyncio Protocol with line parsing and queuing writes"""
 
-    # pylint: disable=too-many-instance-attributes
-    def __init__(self, loop, connected, disconnected, got_data, timeout):
+    # pylint: disable=too-many-instance-attributes, too-many-arguments
+    def __init__(self, loop, heartbeat, connected, disconnected, got_data, timeout):
         self.loop = loop
+        self._heartbeat_time = heartbeat
         self._connected_callback = connected
         self._disconnected_callback = disconnected
         self._got_data_callback = got_data
@@ -46,7 +48,8 @@ class Connection(asyncio.Protocol):
 
         self._transport = None
         self._awaiting = NOTHING
-        self._timeout_task = None
+        self._pim_timeout_task = None
+        self._heartbeat_timeout_task = None
         self._write_queue = []
         self._buffer = ""
         self._paused = False
@@ -54,7 +57,7 @@ class Connection(asyncio.Protocol):
     def connection_made(self, transport):
         LOG.debug("connected callback")
         self._transport = transport
-        self._connected_callback(transport, self)
+        self._connected_callback(self)
 
     def connection_lost(self, exc):
         LOG.debug("disconnected callback")
@@ -63,11 +66,12 @@ class Connection(asyncio.Protocol):
         if self._disconnected_callback:
             self._disconnected_callback()
 
-    def stop(self):
+    def close(self):
         """Stop the connection from sending/receiving/reconnecting."""
-        self._transport = None
+        if self._transport:
+            self._transport.close()
+            self._transport = None
         self._cleanup()
-        self._disconnected_callback = None
 
     def pause(self):
         """Pause the connection from sending/receiving."""
@@ -75,6 +79,7 @@ class Connection(asyncio.Protocol):
         self._paused = True
 
     def is_paused(self):
+        """Is the connections paused?"""
         return self._paused
 
     def resume(self):
@@ -92,14 +97,16 @@ class Connection(asyncio.Protocol):
     def _response_timeout(self):
         kind = "PIM" if self._awaiting == PIM_RESPONSE_MESSAGE else "packet"
         self._awaiting = NOTHING
-        self._timeout_task = None
+        self._pim_timeout_task = None
 
         if not self._write_queue:
-            LOG.warning(f"_response_timeout: No writes are queued.")
+            LOG.warning("_response_timeout: No writes are queued.")
             return
 
         pkt = self._write_queue[0]
-        LOG.debug(f"Timeout waiting for {kind}; retrying {pkt.retry_count} more times.")
+        LOG.debug(
+            "Timeout waiting for %s; retrying %d more times.", kind, pkt.retry_count
+        )
         if pkt.retry_count == 0:
             self._write_queue.pop(0)
             self._timeout_callback(kind, pkt.response)
@@ -151,13 +158,29 @@ class Connection(asyncio.Protocol):
 
         self._got_data_callback(msg)
 
+    def _cancel_heartbeat_timer(self):
+        if self._heartbeat_timeout_task:
+            self._heartbeat_timeout_task.cancel()
+
+    def _heartbeat_timeout(self):
+        LOG.warning("UPB connection heartbeat timed out, disconnecting")
+        self._transport.close()
+
+    def _restart_heartbeat_timer(self):
+        self._cancel_heartbeat_timer()
+        if self._heartbeat_time > 0:
+            self._heartbeat_timeout_task = self.loop.call_later(
+                self._heartbeat_time, self._heartbeat_timeout
+            )
+
     def data_received(self, data):
+        self._restart_heartbeat_timer()
         self._buffer += data.decode("ISO-8859-1")
         pim_busy = False
         while "\r" in self._buffer:
             line, self._buffer = self._buffer.split("\r", 1)
             pim_command = line[:2]
-            LOG.debug(f"data_received:  '{line}'")
+            LOG.debug("data_received:  '%s'", line)
 
             if pim_command == "PU":
                 self._handle_pim_update_msg(line)
@@ -173,8 +196,8 @@ class Connection(asyncio.Protocol):
                 self._got_data_callback(line)
 
         if pim_busy:
-            LOG.debug(f"PIM busy received, retrying in {PIM_BUSY_TIMEOUT} seconds")
-            self._start_timer(PIM_BUSY_TIMEOUT, self._pim_busy_timeout)
+            LOG.debug("PIM busy received, retrying in %s seconds", PIM_BUSY_TIMEOUT)
+            self._start_pim_timer(PIM_BUSY_TIMEOUT, self._pim_busy_timeout)
             self._awaiting = PIM_TO_BE_READY
         else:
             self._process_write_queue()
@@ -186,29 +209,31 @@ class Connection(asyncio.Protocol):
             return
 
         pkt = self._write_queue[0]
-        self._start_timer(pkt.timeout, self._response_timeout)
+        self._start_pim_timer(pkt.timeout, self._response_timeout)
         self._awaiting = UPB_PACKET if pkt.response else PIM_RESPONSE_MESSAGE
 
-        LOG.debug(f"write_data '{pkt.data}'")
+        LOG.debug("write_data '%s'", pkt.data)
         pim_command = "" if pkt.raw else PimCommand.TX_UPB_MSG.value
         self._transport.write(f"{pim_command}{pkt.data}\r".encode())
 
     def _done_with_write_queue_head(self):
-        self._cancel_timer()
+        self._cancel_pim_timer()
         self._awaiting = NOTHING
         if self._write_queue:
             self._write_queue.pop(0)
 
-    def _cancel_timer(self):
-        if self._timeout_task:
-            self._timeout_task.cancel()
-            self._timeout_task = None
+    def _cancel_pim_timer(self):
+        if self._pim_timeout_task:
+            self._pim_timeout_task.cancel()
+            self._pim_timeout_task = None
 
-    def _start_timer(self, timeout, callback):
-        self._cancel_timer()
-        self._timeout_task = self.loop.call_later(timeout, callback)
+    def _start_pim_timer(self, timeout, callback):
+        self._cancel_pim_timer()
+        self._pim_timeout_task = self.loop.call_later(timeout, callback)
 
     def _cleanup(self):
-        self._cancel_timer()
+        self._cancel_pim_timer()
+        self._cancel_heartbeat_timer()
         self._write_queue = []
         self._buffer = ""
+        self._awaiting = NOTHING
