@@ -8,7 +8,7 @@ import serial_asyncio
 
 from .devices import UpbAddr, UpbDevices
 from .links import Links
-from .message import MessageDecode
+from .message import MessageDecode, MessageEncode
 from .parse_upstart import process_upstart_file
 from .proto import Connection
 from .util import parse_flags, parse_url
@@ -19,42 +19,43 @@ LOG = logging.getLogger(__name__)
 class UpbPim:
     """Represents all the components on an UPB PIM."""
 
+    # pylint: disable=too-many-instance-attributes
     def __init__(self, config, loop=None):
         """Initialize a new PIM instance."""
+        self.flags = parse_flags(config.get("flags", ""))
+        LOG.info("Using flags: %s", str(self.flags))
+        self._decoder = MessageDecode()
+        self.encoder = MessageEncode(self.flags.get("tx_count", 1))
+
         self.loop = loop if loop else asyncio.get_event_loop()
         self._config = config
-        self._conn = None
-        self._transport = None
-        self.connection_lost_callbk = None
-        self.connected_callbk = None
-        self._connection_retry_timer = 1
-        self._connection_retry_task = None
-        self._message_decode = MessageDecode()
+        self._connection = None
+        self._connection_retry_time = 1
+        self._reconnect_task = None
         self._sync_handlers = []
-        self._heartbeat = None
-        self.flags = {}
+
         self.devices = UpbDevices(self)
         self.links = Links(self)
         self.config_ok = True
         self.network_id = None
-
-        self.flags = parse_flags(config.get("flags", ""))
+        self.connected_callbk = None
 
         # Setup for all the types of elements tracked
         export_filepath = config.get("UPStartExportFile")
         if export_filepath:
             self.config_ok = process_upstart_file(self, config["UPStartExportFile"])
 
-    async def _connect(self, connected_callbk=None, connection_lost_callbk=None):
+    async def _connect(self, connected_callbk=None):
         """Asyncio connection to UPB."""
         self.connected_callbk = connected_callbk
-        self.connection_lost_callbk = connection_lost_callbk
         url = self._config["url"]
         LOG.info("Connecting to UPB PIM at %s", url)
         scheme, dest, param = parse_url(url)
+        heartbeat_time = self.flags.get("heartbeat_timeout_sec", 90) if scheme == "tcp" else -1
         conn = partial(
             Connection,
             self.loop,
+            heartbeat_time,
             self._connected,
             self._disconnected,
             self._got_data,
@@ -74,27 +75,24 @@ class UpbPim:
             LOG.warning(
                 "Could not connect to UPB PIM (%s). Retrying in %d seconds",
                 err,
-                self._connection_retry_timer,
-            )
-            self._connection_retry_task = self.loop.call_later(
-                self._connection_retry_timer, self.connect
-            )
-            self._connection_retry_timer = (
-                2 * self._connection_retry_timer
-                if self._connection_retry_timer < 32
-                else 60
+                self._connection_retry_time,
             )
+            self._start_connection_retry_timer()
 
-    def _connected(self, transport, conn):
+    def _start_connection_retry_timer(self):
+        timeout = self._connection_retry_time
+        if timeout > 0:
+            self._reconnect_task = self.loop.call_later(timeout, self.connect)
+            self._connection_retry_time = min(timeout * 2, 60)
+
+    def _connected(self, conn):
         """Sync the UPB PIM network to memory."""
         LOG.info("Connected to UPB PIM")
-        self._conn = conn
-        self._transport = transport
-        self._connection_retry_timer = 1
+        self._connection = conn
+        self._connection_retry_time = 1
         if self.connected_callbk:
             self.connected_callbk()
-        self.devices.connection_status_change("connected")
-        self.links.connection_status_change("connected")
+        self._connection_status_change("connected")
 
         # The intention of this is to clear anything in the PIM receive buffer.
         # A number of times on startup error(s) (PE) are returned. This too will
@@ -106,60 +104,57 @@ class UpbPim:
         else:
             self.call_sync_handlers()
 
-    def _reset_connection(self):
-        LOG.warning("PIM connection heartbeat timed out, disconnecting")
-        self._transport.close()
-        self._heartbeat = None
-
     def _disconnected(self):
         LOG.warning("PIM at %s disconnected", self._config["url"])
-        self._conn = None
-        self.devices.connection_status_change("disconnected")
-        self.links.connection_status_change("disconnected")
-        if self.connection_lost_callbk:
-            self.connection_lost_callbk()
-        if self._heartbeat:
-            self._heartbeat.cancel()
-            self._heartbeat = None
-        if not self._disconnect_requested:
-            self._connection_retry_task = self.loop.call_later(
-                self._connection_retry_timer, self.connect
-            )
+        self._connection = None
+        self._connection_status_change("disconnected")
+        self._start_connection_retry_timer()
 
     def add_handler(self, msg_type, handler):
-        self._message_decode.add_handler(msg_type, handler)
+        """Add handler for a message type."""
+        self._decoder.add_handler(msg_type, handler)
 
     def _got_data(self, data):  # pylint: disable=no-self-use
         try:
             if data[:2] == "~~":
-                if data == "~~PAUSE":
-                    LOG.info("PAUSE connection")
-                    self.pause()
-                elif data == "~~RESUME":
-                    LOG.info("RESUME connection")
-                    self.resume()
-                    self.call_sync_handlers()
-                elif data == "~~HEARTBEAT":
-                    if self._heartbeat:
-                        self._heartbeat.cancel()
-                    self._heartbeat = self.loop.call_later(90, self._reset_connection)
+                self._handle_control_command(data)
             else:
-                self._message_decode.decode(data)
+                self._decoder.decode(data)
         except (ValueError, AttributeError) as err:
             LOG.debug(err)
 
+    def _handle_control_command(self, data):
+        log_msg = data[2:].lower().replace("_", " ")
+        if data == "~~PAUSE" or data.startswith("~~SERIAL_DISCONNECTED"):
+            LOG.info("Pausing connection (%s)", log_msg)
+            self.pause()
+        elif data in ["~~RESUME", "~~SERIAL_CONNECTED"]:
+            LOG.info("Resuming connection (%s)", log_msg)
+            self.resume()
+            self.call_sync_handlers()
+        elif data == "~~ANOTHER_TCP_CLIENT_IS_CONNECTED":
+            self._connection_retry_time = 60
+            LOG.warning("%s to ser2tcp, disconnecting", log_msg)
+
+    def _connection_status_change(self, status):
+        self.devices.connection_status_change(status)
+        self.links.connection_status_change(status)
+        self._decoder.call_handlers(status, {})
+
     def _timeout(self, kind, addr):
         if kind == "PIM":
-            LOG.warning(f"Timeout communicating with PIM, is it connected?")
+            LOG.warning("Timeout communicating with PIM, is it connected?")
         else:
             device_id = UpbAddr(int(addr[0:2], 16), int(addr[2:4], 16), 0).index
             device = self.devices.elements.get(device_id)
             if device:
                 LOG.warning(
-                    f"Timeout communicating with UPB device '{device.name}' ({device_id})"
+                    "Timeout communicating with UPB device '%s' (%s)",
+                    device.name,
+                    device_id,
                 )
             else:
-                LOG.warning(f"Timeout communicating with UPB device {device_id}")
+                LOG.warning("Timeout communicating with UPB device %s", device_id)
 
     def add_sync_handler(self, sync_handler):
         """Register a fn that synchronizes elements."""
@@ -173,28 +168,21 @@ class UpbPim:
 
     def is_connected(self):
         """Status of connection to PIM."""
-        return self._conn is not None and not self._conn.is_paused()
+        return self._connection is not None and not self._connection.is_paused()
 
-    def connect(self, connected_callbk=None, connection_lost_callbk=None):
+    def connect(self, connected_callbk=None):
         """Connect to the panel"""
-        self._disconnect_requested = False
-        asyncio.ensure_future(self._connect(connected_callbk, connection_lost_callbk))
+        asyncio.ensure_future(self._connect(connected_callbk))
 
     def disconnect(self):
         """Disconnect the connection from sending/receiving."""
-        self._disconnect_requested = True
-        if self._conn:
-            self._conn.stop()
-            self._conn = None
-        if self._connection_retry_task:
-            self._connection_retry_task.cancel()
-            self._connection_retry_task = None
-        if self._transport:
-            self._transport.close()
-            self._transport = None
-        if self._heartbeat:
-            self._heartbeat.cancel()
-            self._heartbeat = None
+        self._connection_retry_time = -1  # Stop from reconnecting automatically
+        if self._connection:
+            self._connection.close()
+            self._connection = None
+        if self._reconnect_task:
+            self._reconnect_task.cancel()
+            self._reconnect_task = None
 
     def run(self):
         """Enter the asyncio loop."""
@@ -202,19 +190,17 @@ class UpbPim:
 
     def send(self, msg, response_required=True, raw=False):
         """Send a message to UPB PIM."""
-        if self._conn:
-            self._conn.write_data(msg, response_required, raw=raw)
+        if self._connection:
+            self._connection.write_data(msg, response_required, raw=raw)
 
     def pause(self):
         """Pause the connection from sending/receiving."""
-        if self._conn:
-            self.devices.connection_status_change("paused")
-            self.links.connection_status_change("paused")
-            self._conn.pause()
+        if self._connection:
+            self._connection_status_change("paused")
+            self._connection.pause()
 
     def resume(self):
         """Restart the connection from sending/receiving."""
-        if self._conn:
-            self.devices.connection_status_change("resume")
-            self.links.connection_status_change("resume")
-            self._conn.resume()
+        if self._connection:
+            self._connection_status_change("resume")
+            self._connection.resume()
